@@ -1,3 +1,4 @@
+//go:build gc.none
 // +build gc.none
 
 package runtime
@@ -10,8 +11,12 @@ import (
 	"unsafe"
 )
 
+const gcAsserts = false // perform sanity checks
+
 func alloc(size uintptr, layout unsafe.Pointer) unsafe.Pointer
 
+func realloc(ptr unsafe.Pointer, size uintptr) unsafe.Pointer
+
 func free(ptr unsafe.Pointer) {
 	// Nothing to free when nothing gets allocated.
 }
@@ -35,3 +40,7 @@ func initHeap() {
 func setHeapEnd(newHeapEnd uintptr) {
 	// Nothing to do here, this function is never actually called.
 }
+
+func markRoots(start, end uintptr) {
+	// dummy, so that markGlobals will compile
+}
