@@ -1,3 +1,4 @@
+//go:build rp2040
 // +build rp2040
 
 package runtime
@@ -11,6 +12,9 @@ import (
 // machineTicks is provided by package machine.
 func machineTicks() uint64
 
+// machineLightSleep is provided by package machine.
+func machineLightSleep(uint64)
+
 type timeUnit uint64
 
 // ticks returns the number of ticks (microseconds) elapsed since power up.
@@ -31,6 +35,16 @@ func sleepTicks(d timeUnit) {
 	if d == 0 {
 		return
 	}
+
+	if hasScheduler {
+		// With scheduler, sleepTicks may return early if an interrupt or
+		// event fires - so scheduler can schedule any go routines now
+		// eligible to run
+		machineLightSleep(uint64(d))
+		return
+	}
+
+	// Busy loop
 	sleepUntil := ticks() + d
 	for ticks() < sleepUntil {
 	}
@@ -44,6 +58,18 @@ func putchar(c byte) {
 	machine.Serial.WriteByte(c)
 }
 
+func getchar() byte {
+	for machine.Serial.Buffered() == 0 {
+		Gosched()
+	}
+	v, _ := machine.Serial.ReadByte()
+	return v
+}
+
+func buffered() int {
+	return machine.Serial.Buffered()
+}
+
 // machineInit is provided by package machine.
 func machineInit()
 
